import logging
from django.contrib.auth import authenticate, logout
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied
# from django.db.models import Q
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from .auth import login, PLATFORM_SESSION_KEY
from .decorators import Admin_login_required, User_login_required
from .forms import LoginForm, UserProfileEditForm, UserProfileRecoveryForm
from .helper import (
    get_all_platform_names_from_DB, get_form_error_messages, _display_user_profile_edit_error_messages,
    _display_user_profile_recovery_error_messages, _display_user_profile_recovery_success_message, _update_user_profile
)
from .models import Registry, User

# Create your views here.
# https://docs.djangoproject.com/en/2.0/topics/http/views/

# For more information on configuring and using the logging system, see https://docs.djangoproject.com/en/2.0/topics/logging/ and the "Logging" part of settings.py .
# For more information on default/customizing user authentication and password management, see https://docs.djangoproject.com/en/2.0/topics/auth/, auth.py, backends.py and AuthenticationMiddleware of middleware.py .
# For more information on making database queries, i.e. creating, retrieving, updating and deleting model instances, see https://docs.djangoproject.com/en/2.0/topics/db/queries/ .
# For more information on working with forms, see https://docs.djangoproject.com/en/2.0/topics/forms/ .
# For more information on using TemplateResponse objects - instead of standard HttpResponse objects and calls to the render() shortcut function - to display application pages/templates, see https://docs.djangoproject.com/en/2.0/ref/template-response/ .

# The below defined views refer to the application site; therefore, they are not targeted at the default user model but at the application user models (see the _get_user_model_label() function of auth.py), i.e. the views can be accessed only by usermerge.{User/Admin} instances (the "usermerge." prefix is usually omitted in the application context) and not by auth.User ones - see the User_login_required() and Admin_login_required() decorators. Furthermore, the user instances (including the session one) that are created, retrieved, updated or deleted in the context of each view are usermerge.{User/Admin} instances. On the other hand, the admin site (see https://docs.djangoproject.com/en/2.0/ref/contrib/admin/) is not targeted at the application user models but at the default user model as only auth.User instances can be active staff members - each one has the is_active and is_staff attributes and the latter are set to True - and also have permissions - each one has the is_superuser, groups and user_permissions attributes as well as the relative methods, e.g. get_group_permissions(), get_all_permissions(), etc. - to access the site and utilize some or all of its capabilities. The admin site should generally be used for retrieving/reviewing model instances and not for creating, updating or deleting them - the latter actions should better be performed by appropriately defined views. When trying to create, update or delete instances via the admin site, extra care should be given to retain consistency in usermergeDB, e.g. a Platform instance could be created without any problem, while the updated password of an Admin instance would wrongly be stored in raw format (without being hashed first), etc. (for more information, see models.py).
# For more information on the possible/valid (field) states of a user profile (User instance) in usermergeDB and the allowed/consistent transitions between them, see models.py . It is worth mentioning that a user profile is empty if its email is None and vice versa - see the has_empty_profile() method of User model.
# A user whose model is User and profile is non-empty should have access to all the User-targeted, i.e. User_login_required()-decorated, views except for the user profile recovery ones. On the other hand, a user whose model is User and profile is empty should have access only to the User-targeted views that relate to the user home page and the user profile edit/recovery. Therefore, he/she should fill out his/her profile or recover a non-empty one from usermergeDB first to gain access to all the other User-targeted views (these views should ensure that only non-empty user profiles are allowed access) - if this was done, he/she would have access to all the User-targeted views except for the user profile recovery ones. Finally, a user whose model is Admin should have access to all the Admin-targeted, i.e. Admin_login_required()-decorated, views.

def display_default_login_page(request): # The default login page is the index page of the application site.
    """
    Check if the current user has logged in. If he/she has not (his/her model is AnonymousUser), display the login page in default
    mode, i.e. without any form-submission-related messages. Otherwise, check if his/her model is auth.User . If it is (he/she has
    logged in to the admin site), display an HTTP 403 (Forbidden) page by raising a PermissionDenied exception (he/she has logged
    in to the admin site and should thus be prevented from accessing any application page/template). Otherwise (his/her model is
    usermerge.{User/Admin} and he/she has logged in to the application site), display the session continuation page which informs
    him/her that his/her session is still active and provides him/her with the choices of either navigating to his/her model-specific
    home page (if he/she would like to continue navigating in this session) or logging out of the application site (if he/she would
    like to start a new session as another user).
    """
    current_user = request.user
    if current_user.is_authenticated:
        current_user_model_label = current_user._meta.label
        if current_user_model_label == 'usermerge.User':
            return TemplateResponse(request, 'session_continuation.html', {'home_page_url': 'user_home'})
        elif current_user_model_label == 'usermerge.Admin':
            return TemplateResponse(request, 'session_continuation.html', {'home_page_url': 'admin_home'})
        else: # current_user_model_label == 'auth.User'
            raise PermissionDenied
    else: # current user model is AnonymousUser
        return TemplateResponse(request, 'login.html', {'platform_names': get_all_platform_names_from_DB()})

def log_in(request):
    """
    Check if the current user has logged in. If he/she has (he/she has logged in either to the admin site as an auth.User instance or
    to the application site as a usermerge.{User/Admin} instance), display an HTTP 403 (Forbidden) page by raising a PermissionDenied
    exception (he/she has logged in either to the admin site and should thus be prevented from accessing any application view or to
    the application site and should thus be prevented from reattempting to log in to the latter). Otherwise (he/she is an AnonymousUser
    instance), collect the data, i.e. platform, username and password, POSTed via the login form and validate them. If they are not
    valid, display the appropriate validation error messages in the login page. Otherwise, authenticate the credentials (username and
    password) for the selected platform. If they do not specify any usermerge.{User/Admin} instance of usermergeDB, display the
    appropriate post-validation error message in the login page and include the corresponding error code in the page context. Otherwise,
    log the current user in to the application site as the specified (authenticated) usermerge.{User/Admin} instance and redirect him/her
    to his/her model-specific home page.
    """
    if request.user.is_authenticated: # The current user is either an auth.User or a usermerge.{User/Admin} instance.
        raise PermissionDenied
    else: # The current user is an AnonymousUser instance.
        form = LoginForm(request.POST)
        if form.is_valid():
            platform = form.cleaned_data['platform']
            username = form.cleaned_data['username']
            password = form.cleaned_data['password']
            auth_user = authenticate(request, platform = platform, username = username, password = password)
            if auth_user:
                login(request, auth_user)
                # If the selected platform is SLUB, the authenticated (just logged-in) user's model is usermerge.Admin .
                # Otherwise, it is usermerge.User .
                if platform == 'SLUB':
                    return redirect('admin_home')
                else:
                    return redirect('user_home')
            else:
                return TemplateResponse(request, 'login.html',
                                        {'platform_names': get_all_platform_names_from_DB(),
                                         'error_messages': ['Τα δηλωθέντα στοιχεία εισόδου δεν ανταποκρίνονται '
                                                            'σε κανέναν χρήστη της βάσης!\n'
                                                            'Παρακαλούμε ελέγξτε την ορθότητα των στοιχείων εισόδου και δοκιμάστε ξανά!'],
                                         'post_validation_error_codes': ['non_existent_credentials_pair_for_selected_login_platform']})
        else:
            return TemplateResponse(request, 'login.html', {'platform_names': get_all_platform_names_from_DB(),
                                                            'error_messages': get_form_error_messages(form)})

@User_login_required
def display_user_home_page(request):
    """
    Display the user home page.
    """
    return TemplateResponse(request, 'user_home.html', {})

@User_login_required
def display_default_user_profile_edit_page(request):
    """
    Display the user profile edit page in default mode, i.e. without any form-submission-related messages.
    """
    return TemplateResponse(request, 'user_profile_edit.html', {})

@User_login_required
def edit_user_profile(request):
    """
    Collect the data, i.e. first name, last name, ece_id and email, POSTed via the user profile edit form and validate them. If they
    are not valid, display the appropriate validation error messages in the user profile edit page. Otherwise, check if any form fields
    have been changed in comparison to the respective fields of the session user's profile. If no changes have been made, display the
    appropriate post-validation error message in the user profile edit page and include the corresponding error code in the page context.
    Otherwise, check if the validated (cleaned) ece_id and email are adequate to modify the session user's profile consistently in
    usermergeDB by taking the respective field values of the latter into account. If they are not adequate, create a dictionary with
    them as values and the corresponding (post-validation) error codes as keys and call the _display_user_profile_edit_error_messages()
    helper function with the aforementioned dictionary as the error_data argument. Otherwise, call the _update_user_profile() helper
    function with the (name) list of all the changed form fields as the changed_data argument and the dictionary of all the validated
    form data (the field names are used as keys and the corresponding validated values as values) as the cleaned_data argument.
    """
    session_user = request.user
    form = UserProfileEditForm(request.POST, initial = {'first_name': session_user.first_name, 'last_name': session_user.last_name,
                                                        'ece_id': session_user.ece_id, 'email': session_user.email})
    if form.is_valid():
        cleaned_data = form.cleaned_data
        # first_name = cleaned_data['first_name']
        # last_name = cleaned_data['last_name']
        ece_id = cleaned_data['ece_id']
        email = cleaned_data['email']
        if not form.has_changed():
            return TemplateResponse(request, 'user_profile_edit.html',
                                    {'error_messages': ['Η βάση δεν χρειάστηκε να ενημερωθεί καθώς κανένα από τα δηλωθέντα στοιχεία του '
                                                        'προφίλ δεν τροποποιήθηκε!\n'
                                                        'Παρακαλούμε, προτού πατήσετε "Αποθήκευση", βεβαιωθείτε πρώτα ότι έχετε '
                                                        'τροποποιήσει ένα ή περισσότερα\n'
                                                        'στοιχεία του προφίλ!'],
                                     'post_validation_error_codes': ['unchanged_profile_form_fields']})
        else:
            changed_data = form.changed_data
            # The session user's profile in usermergeDB is empty (session_user.ece_id is None and session_user.email is None).
            if session_user.has_empty_profile():
                try:
                    User.objects.get(email = email)
                except User.DoesNotExist:
                    try:
                        User.objects.get(ece_id = ece_id)
                    # The validated ece_id either exists multiple times (ece_id is None) or
                    # does not exist at all (ece_id is not None) in usermergeDB.
                    except (User.DoesNotExist, User.MultipleObjectsReturned):
                        return _update_user_profile(request, changed_data, cleaned_data)
                    else:
                        if ece_id is None:
                            return _update_user_profile(request, changed_data, cleaned_data)
                        else:
                            return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id})
                else:
                    try:
                        User.objects.get(ece_id = ece_id)
                    # The validated ece_id either exists multiple times (ece_id is None) or
                    # does not exist at all (ece_id is not None) in usermergeDB.
                    except (User.DoesNotExist, User.MultipleObjectsReturned):
                        return _display_user_profile_edit_error_messages(request, {'duplicate_email': email})
                    else:
                        if ece_id is None:
                            return _display_user_profile_edit_error_messages(request, {'duplicate_email': email})
                        else:
                            return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id,
                                                                                       'duplicate_email': email})
            else: # The session user's profile in usermergeDB is non-empty (session_user.email is not None).
                if session_user.ece_id is None:
                    if ece_id is None:
                        if 'email' in changed_data:
                            try:
                                User.objects.get(email = email)
                            except User.DoesNotExist:
                                return _update_user_profile(request, changed_data, cleaned_data)
                            else:
                                return _display_user_profile_edit_error_messages(request, {'duplicate_email': email})
                        else:
                            return _update_user_profile(request, changed_data, cleaned_data)
                    else: # ece_id is not None
                        try:
                            User.objects.get(ece_id = ece_id)
                        except User.DoesNotExist:
                            if 'email' in changed_data:
                                try:
                                    User.objects.get(email = email)
                                except User.DoesNotExist:
                                    return _update_user_profile(request, changed_data, cleaned_data)
                                else:
                                    return _display_user_profile_edit_error_messages(request, {'duplicate_email': email})
                            else:
                                return _update_user_profile(request, changed_data, cleaned_data)
                        else:
                            if 'email' in changed_data:
                                try:
                                    User.objects.get(email = email)
                                except User.DoesNotExist:
                                    return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id})
                                else:
                                    return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id,
                                                                                               'duplicate_email': email})
                            else:
                                return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id})
                else: # session_user.ece_id is not None
                    if ece_id is None:
                        if 'email' in changed_data:
                            try:
                                User.objects.get(email = email)
                            except User.DoesNotExist:
                                return _display_user_profile_edit_error_messages(request,
                                                                                 {'registered_ece_id_was_cleared': session_user.ece_id})
                            else:
                                return _display_user_profile_edit_error_messages(request,
                                                                                 {'registered_ece_id_was_cleared': session_user.ece_id,
                                                                                  'duplicate_email': email})
                        else:
                            return _display_user_profile_edit_error_messages(request,
                                                                             {'registered_ece_id_was_cleared': session_user.ece_id})
                    else: # ece_id is not None
                        if 'ece_id' in changed_data:
                            try:
                                User.objects.get(ece_id = ece_id)
                            except User.DoesNotExist:
                                if 'email' in changed_data:
                                    try:
                                        User.objects.get(email = email)
                                    except User.DoesNotExist:
                                        return _update_user_profile(request, changed_data, cleaned_data)
                                    else:
                                        return _display_user_profile_edit_error_messages(request, {'duplicate_email': email})
                                else:
                                    return _update_user_profile(request, changed_data, cleaned_data)
                            else:
                                if 'email' in changed_data:
                                    try:
                                        User.objects.get(email = email)
                                    except User.DoesNotExist:
                                        return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id})
                                    else:
                                        return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id,
                                                                                                   'duplicate_email': email})
                                else:
                                    return _display_user_profile_edit_error_messages(request, {'duplicate_ece_id': ece_id})
                        else:
                            if 'email' in changed_data:
                                try:
                                    User.objects.get(email = email)
                                except User.DoesNotExist:
                                    return _update_user_profile(request, changed_data, cleaned_data)
                                else:
                                    return _display_user_profile_edit_error_messages(request, {'duplicate_email': email})
                            else:
                                return _update_user_profile(request, changed_data, cleaned_data)
    else: # not form.is_valid()
        return TemplateResponse(request, 'user_profile_edit.html', {'error_messages': get_form_error_messages(form)})

@User_login_required
def display_default_user_profile_recovery_page(request):
    """
    Check if the session user's profile in usermergeDB is empty. If it is not, display an HTTP 403 (Forbidden) page by raising a
    PermissionDenied exception (there is no point in specifying and recovering a non-empty user profile from usermergeDB for the
    session user if the latter's profile is already non-empty - see the search_for_recovery_user_profile() and recover_user_profile()
    views below). Otherwise, display the user profile recovery page in default mode, i.e. without any form-submission-related messages.
    """
    session_user = request.user
    if not session_user.has_empty_profile():
        raise PermissionDenied
    else:
        return TemplateResponse(request, 'user_profile_recovery.html', {})

@User_login_required
def search_for_recovery_user_profile(request):
    """
    Check if the session user's profile in usermergeDB is empty. If it is not, display an HTTP 403 (Forbidden) page by raising a
    PermissionDenied exception (there is no point in searching usermergeDB to recover a non-empty profile for the session user if
    the latter's profile is already non-empty). Otherwise, collect the data, i.e. ece_id and email, POSTed via the user profile
    recovery form and validate them. If they are not valid, display the appropriate validation error messages in the user profile
    recovery page. Otherwise, check if they are adequate to specify a non-empty user profile (User instance) for recovery in
    usermergeDB (the specified profile would correspond to both the validated - cleaned - ece_id and email). If they are not
    adequate (the validated ece_id or/and email do not exist in usermergeDB or they both exist in the latter but not in the same
    profile), create a dictionary with them as values and the corresponding (post-validation) error codes as keys and call the
    _display_user_profile_recovery_error_messages() helper function with the aforementioned dictionary as the error_data argument.
    Otherwise, call the _display_user_profile_recovery_success_message() helper function with the specified profile as the
    recov_user argument.
    """
    # The current view serves as the preparatory step of the user profile recovery procedure, i.e. it is used to set up the
    # corresponding session and page contexts appropriately before the recovery procedure takes place - see the recover_user_profile()
    # view below. Both contexts are set up by the called helper function - either _display_user_profile_recovery_success_message()
    # or _display_user_profile_recovery_error_messages(). For the session one in particular, the state of the recov_user_id value is
    # appropriately modified in the session data to represent the latest search result of the current view, i.e. either the
    # recov_user_id value exists in the session data and represents the id of the specified recovery user profile or it does not
    # exist and indicates that no recovery user profile has been specified. The recov_user_id value is stored in the session data and
    # is obtained from there by the recover_user_profile() view instead of being given to that view as a URL argument because in the
    # latter case it would be susceptible to manual change (e.g. it could be misedited via JavaScript) and could thus bypass the
    # appropriate recovery user profile specification provided by the current view.
    session_user = request.user
    # The session user's profile in usermergeDB is non-empty (session_user.email is not None).
    if not session_user.has_empty_profile():
        raise PermissionDenied
    else: # The session user's profile in usermergeDB is empty (session_user.ece_id is None and session_user.email is None).
        form = UserProfileRecoveryForm(request.POST)
        if form.is_valid():
            ece_id = form.cleaned_data['ece_id']
            email = form.cleaned_data['email']
            # Taking into account that the validated email is never None (the validated ece_id can be None), the happy path of the view
            # is the following:
            # * Search usermergeDB for a non-empty user profile that corresponds to the validated email (in other words, check if the
            #   validated email exists in usermergeDB).
            # * If a corresponding profile is found, check if the validated ece_id matches the one of the profile (in other words, check
            #   if the validated ece_id exists in usermergeDB and specifically in the same profile as the validated email).
            # * If the validated ece_id matches the one of the profile (the profile corresponds to both the validated ece_id and email),
            #   the search succeeds, the profile is deemed the appropriate one for recovery and the
            #   _display_user_profile_recovery_success_message() function is called.
            # If any of the checks (searching for the corresponding profile and comparing the ece_id values) fails, the search fails,
            # the exact search errors (error codes) are determined by the existence statuses of both the validated ece_id and email
            # in usermergeDB and the _display_user_profile_recovery_error_messages() function is called (if the first check fails,
            # the second one is never performed).
            try:
                recov_user = User.objects.get(email = email)
            except User.DoesNotExist:
                try:
                    User.objects.get(ece_id = ece_id)
                except User.DoesNotExist: # The validated ece_id does not exist in usermergeDB (ece_id is not None).
                    return _display_user_profile_recovery_error_messages(request, {'non_existent_ece_id': ece_id,
                                                                                   'non_existent_email': email})
                except User.MultipleObjectsReturned: # The validated ece_id exists multiple times in usermergeDB (ece_id is None).
                    try:
                        # Taking into account that empty user profiles (email is None) are indifferent to the search procedure and that
                        # the validated ece_id is None, search usermergeDB for non-empty user profiles (email is not None) whose ece_id
                        # is None (None is interpreted as NULL in usermergeDB) and specify the exact error_data argument of the called
                        # _display_user_profile_recovery_error_messages() function based on the search result.
                        User.objects.get(ece_id__isnull = True, email__isnull = False)
                    except User.DoesNotExist:
                        return _display_user_profile_recovery_error_messages(request, {'non_existent_ece_id': ece_id,
                                                                                       'non_existent_email': email})
                    except User.MultipleObjectsReturned:
                        return _display_user_profile_recovery_error_messages(request, {'non_existent_email': email})
                    else:
                        return _display_user_profile_recovery_error_messages(request, {'non_existent_email': email})
                else:
                    if ece_id is None:
                        return _display_user_profile_recovery_error_messages(request, {'non_existent_ece_id': ece_id,
                                                                                       'non_existent_email': email})
                    else:
                        return _display_user_profile_recovery_error_messages(request, {'non_existent_email': email})
            else: # The validated email exists exactly once in usermergeDB (recov_user.email == email).
                if recov_user.ece_id is None:
                    if ece_id is None:
                        return _display_user_profile_recovery_success_message(request, recov_user)
                    else:
                        try:
                            User.objects.get(ece_id = ece_id)
                        except User.DoesNotExist:
                            return _display_user_profile_recovery_error_messages(request, {'non_existent_ece_id': ece_id})
                        else:
                            return _display_user_profile_recovery_error_messages(request, {'ece_id_and_email_exist_in_different_profiles':
                                                                                           {'ece_id': ece_id, 'email': email}})
                else: # recov_user.ece_id is not None
                    if ece_id is None:
                        try:
                            # Taking into account that empty user profiles (email is None) are indifferent to the search procedure and
                            # that the validated ece_id is None, search usermergeDB for non-empty user profiles (email is not None) whose
                            # ece_id is None (None is interpreted as NULL in usermergeDB) and specify the exact error_data argument of
                            # the called _display_user_profile_recovery_error_messages() function based on the search result.
                            User.objects.get(ece_id__isnull = True, email__isnull = False)
                        except User.DoesNotExist:
                            return _display_user_profile_recovery_error_messages(request, {'non_existent_ece_id': ece_id})
                        except User.MultipleObjectsReturned:
                            return _display_user_profile_recovery_error_messages(request, {'ece_id_and_email_exist_in_different_profiles':
                                                                                           {'ece_id': ece_id, 'email': email}})
                        else:
                            return _display_user_profile_recovery_error_messages(request, {'ece_id_and_email_exist_in_different_profiles':
                                                                                           {'ece_id': ece_id, 'email': email}})
                    else:
                        if recov_user.ece_id == ece_id:
                            return _display_user_profile_recovery_success_message(request, recov_user)
                        else:
                            try:
                                User.objects.get(ece_id = ece_id)
                            except User.DoesNotExist:
                                return _display_user_profile_recovery_error_messages(request, {'non_existent_ece_id': ece_id})
                            else:
                                return _display_user_profile_recovery_error_messages(request,
                                                                                     {'ece_id_and_email_exist_in_different_profiles':
                                                                                      {'ece_id': ece_id, 'email': email}})
        else: # not form.is_valid()
            return TemplateResponse(request, 'user_profile_recovery.html', {'error_messages': get_form_error_messages(form)})

@User_login_required
def recover_user_profile(request):
    """
    Check if the session user's profile in usermergeDB is empty and if the recov_user_id value exists in the session data, i.e. if a
    non-empty recovery user profile (User instance) has been specified in usermergeDB via the search_for_recovery_user_profile() view.
    If any of the checks fails, display an HTTP 403 (Forbidden) page by raising a PermissionDenied exception (it is impossible to recover
    an unspecified non-empty user profile from usermergeDB and there is no point in recovering a specified one for the session user if
    the latter's profile is already non-empty). Otherwise, delete any previous credentials that the recovery user may have retained for
    the login platform in usermergeDB and associate him/her with the session user's respective ones. More specifically, delete the
    recovery registry, i.e. the Registry instance that corresponds to both the recovery user and the login platform, if it exists in
    usermergeDB and update the existing session registry, i.e. the Registry instance that corresponds to both the session user and the
    login platform, to reference the recovery user instead. Then delete the session user's profile to complete his/her merge with the
    recovery user and if the latter's credentials for the login platform have changed, i.e. he/she retained previous credentials before
    receiving the session user's respective ones, use the view logger (usermerge.views.recover_user_profile) to log the appropriate INFO
    message that mentions the recovery user's previous and current username (key credential) for the login platform. Finally, log the
    session user out (flush the session as it is based on a no longer existing user - profile - and has thus become inconsistent) and
    redirect him/her to the default login page.
    """
    # For more information on the recov_user_id value of the session data, see the search_for_recovery_user_profile() view above.
    view_logger = logging.getLogger('usermerge.views.recover_user_profile')
    session_user = request.user
    if not (session_user.has_empty_profile() and 'recov_user_id' in request.session):
        raise PermissionDenied
    else:
        recov_user_id = request.session['recov_user_id']
        login_platform_name = request.session[PLATFORM_SESSION_KEY]
        session_registry = Registry.objects.get(user = session_user, platform__name = login_platform_name)
        prev_recov_username = ''
        # The changes in usermergeDB should be made in the following order:
        # * Due to the user-platform unique key constraint of the Registry model, the recovery registry should be deleted if it exists
        #   before the session registry is updated to reference the recovery user.
        # * Due to the user foreign key constraint of the Registry model, the session registry should be updated to reference the
        #   recovery user (profile) before the session user's profile is deleted.
        try:
            recov_registry = Registry.objects.get(user_id = recov_user_id, platform__name = login_platform_name)
        except Registry.DoesNotExist:
            pass
        else:
            prev_recov_username = recov_registry.username
            recov_registry.delete()
        session_registry.user_id = recov_user_id
        session_registry.save(update_fields = ['user'])
        session_user.delete()
        if prev_recov_username:
            view_logger.info('user_id: %d | platform: %s (id: %d) | old_username: %s | new_username: %s',
                             recov_user_id, login_platform_name, session_registry.platform_id,
                             prev_recov_username, session_registry.username)
        # On the one hand, the logout() authentication function emits a user_logged_out signal that, among others, provides
        # the corresponding handler functions with the logged-out user's profile. On the other hand, the session user's profile
        # has been deleted from usermergeDB and cannot be accessed by any function. Therefore, instead of calling the log_out()
        # view that utilizes the logout() function, log the session user out (flush the session - delete the session data and
        # cookie - and set an AnonymousUser instance as the current user in the associated request) and redirect him/her to the
        # default login page manually.
        request.session.flush()
        request.user = AnonymousUser()
        return redirect('default_login')

@Admin_login_required
def display_admin_home_page(request):
    """
    Display the admin home page.
    """
    return TemplateResponse(request, 'admin_home.html', {})

# The session user's model can be either auth.User or usermerge.{User/Admin} .
@login_required(redirect_field_name = None, login_url = 'default_login')
def log_out(request):
    """
    Check if the session user's model is auth.User . If it is (he/she has logged in to the admin site), display an HTTP 403 (Forbidden)
    page by raising a PermissionDenied exception (he/she has logged in to the admin site and should thus be prevented from accessing
    any application view). Otherwise (his/her model is usermerge.{User/Admin} and he/she has logged in to the application site), log
    him/her out and redirect him/her to the default login page.
    """
    if request.user._meta.label == 'auth.User':
        raise PermissionDenied
    else: # request.user._meta.label in ['usermerge.User', 'usermerge.Admin']
        logout(request)
        return redirect('default_login')

